@@ -1,5 +1,5 @@
 /************************************************************************************
-   Copyright (C) 2013 SkySQL AB
+   Copyright (C) 2013,2019 MariaDB Corporation AB
    
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
@@ -16,7 +16,6 @@
    or write to the Free Software Foundation, Inc., 
    51 Franklin St., Fifth Floor, Boston, MA 02110, USA
 *************************************************************************************/
-#define WIN32_LEAN_AND_MEAN
 
 #include <Windows.h>
 #include <stdlib.h>
@@ -24,13 +23,15 @@
 #include <tchar.h>
 #include <windowsx.h>
 #include <winuser.h>
+#include <shobjidl.h>
 #include "resource.h"
 #include <ma_odbc.h>
 #include <ma_odbc_setup.h>
 
 #pragma comment(lib, "ComCtl32.lib")
 
-#define LASTPAGE 4
+#define CONNSTR_BUFFER_SIZE 1024
+#define LASTPAGE 5
 
 HINSTANCE     hInstance;
 unsigned int  CurrentPage=  0;
@@ -44,11 +45,19 @@ char          DSNStr[2048];
 HWND          hwndTab[7], hwndMain;
 const int    *EffectiveDisabledPages=    NULL,
              *EffectiveDisabledControls= NULL;
+BOOL          OpenCurSelection=          TRUE;
+
+/* On Windows we are supposed to have schannel, or we can have openssl */
+#if defined(_WIN32) || defined(HAVE_OPENSSL) 
+# define  SSL_DISABLED 0
+#else
+# define  SSL_DISABLED 1
+#endif
 
 const int DisabledPages[MAODBC_PROMPT_REQUIRED + 1][LASTPAGE + 1]= {
-                                                                    { 0, 0, 0, 0, 0},
-                                                                    { 0, 0, 0, 0, 0},
-                                                                    { 1, 0, 1, 1, 1}
+                                                                    { 0, 0, 0, 0, SSL_DISABLED, 0},
+                                                                    { 0, 0, 0, 0, SSL_DISABLED, 0},
+                                                                    { 1, 0, 1, 1, 1, 1}
                                                                    };
 const int   PromptDisabledControls[]= { txtDsnName, 0 },
             EmptyDisabledControls[]=  { 0 };
@@ -61,40 +70,69 @@ const int*  DisabledControls[]=       {
 MADB_DsnMap DsnMap[] = {
   {&DsnKeys[0],  0, txtDsnName,          64, 1},
   {&DsnKeys[1],  0, txtDSNDescription,   64, 0},
-  {&DsnKeys[3],  1, rbPipe,               0, 0},
-  {&DsnKeys[4],  1, rbTCP,                0, 0},
-  {&DsnKeys[5],  1, txtServerName,      128, 0},
-  {&DsnKeys[6],  1, txtUserName,         64, 0},
-  {&DsnKeys[7],  1, txtPassword,         64, 0},
-  {&DsnKeys[8],  1, cbDatabase,           0, 0},
-  {&DsnKeys[9],  1, txtPort,              5, 0},
-  {&DsnKeys[10], 2, txtInitCmd,        2048, 0},
-  {&DsnKeys[11], 2, txtConnectionTimeOut, 5, 0},
-  {&DsnKeys[12], 2, ckReconnect,          0, 0},
-  {&DsnKeys[13], 2, ckConnectPrompt,      0, 0},
-  {&DsnKeys[14], 2, cbCharset,            0, 0},
+  {&DsnKeys[5],  1, rbPipe,               0, 0},
+  {&DsnKeys[6],  1, rbTCP,                0, 0},
+  {&DsnKeys[7],  1, txtServerName,      128, 0},
+  {&DsnKeys[29], 1, txtSocket,          128, 0},
+  {&DsnKeys[8],  1, txtUserName,         64, 0},
+  {&DsnKeys[9],  1, txtPassword,         64, 0},
+  {&DsnKeys[10], 1, cbDatabase,           0, 0},
+  {&DsnKeys[11], 1, txtPort,              5, 0},
+  {&DsnKeys[12], 2, txtInitCmd,        2048, 0},
+  {&DsnKeys[13], 2, txtConnectionTimeOut, 5, 0},
+  {&DsnKeys[39], 2, txtReadTimeOut,       5, 0},
+  {&DsnKeys[40], 2, txtWriteTimeOut,      5, 0},
+  {&DsnKeys[36], 2, cbInteractive,        0, 0},
+  {&DsnKeys[14], 2, ckReconnect,          0, 0},
+  {&DsnKeys[15], 2, ckConnectPrompt,      0, 0},
+  {&DsnKeys[16], 2, cbCharset,            0, 0},
+  {&DsnKeys[34], 2, txtServerKey,       260, 0},
+  {&DsnKeys[18], 3, txtPluginDir,       260, 0},
+  {&DsnKeys[38], 3, ckSchParamNoError,    0, 0},
+  {&DsnKeys[41], 3, ckNoLocalInfile,      0, 0},
+  {&DsnKeys[42], 3, ckNullSchemaMeansCurrent, 0, 0},
+  {&DsnKeys[19], 4, txtSslKey,          260, 0},
+  {&DsnKeys[20], 4, txtSslCert,         260, 0},
+  {&DsnKeys[21], 4, txtSslCertAuth,     260, 0},
+  {&DsnKeys[22], 4, txtSslCaPath,       260, 0},
+  {&DsnKeys[23], 4, txtSslCipher,        32, 0},
+  {&DsnKeys[24], 4, cbSslVerify,          0, 0},
+  {&DsnKeys[32], 4, cbTls11,              1, 0},
+  {&DsnKeys[32], 4, cbTls12,              2, 0},
+  {&DsnKeys[32], 4, cbTls13,              4, 0},
+  {&DsnKeys[33], 4, cbForceTls,           0, 0},
+  {&DsnKeys[27], 4, txtCrl,               0, 0},
+  {&DsnKeys[25], 4, txtTlsPeerFp,        41, 0},
+  {&DsnKeys[26], 4, txtTlsPeerFpList,   260, 0},
   {NULL, 0, 0, 0, 0}
 };
 
+#define CBGROUP_BIT(MapIdx)             (char)(DsnMap[MapIdx].MaxLength)
+#define IS_CB_CHECKED(MapIdx)           GetButtonState(DsnMap[MapIdx].Page, DsnMap[MapIdx].Item)
+#define CBGROUP_SETBIT(_Dsn, MapIdx)    *GET_FIELD_PTR(_Dsn, DsnMap[MapIdx].Key, char)|= CBGROUP_BIT(MapIdx)
+#define CBGROUP_RESETBIT(_Dsn, MapIdx)  *GET_FIELD_PTR(_Dsn, DsnMap[MapIdx].Key, char)&= ~CBGROUP_BIT(MapIdx)
+
 MADB_OptionsMap OptionsMap[]= {
-  {1, rbPipe,                   MADB_OPT_FLAG_NAMED_PIPE},
-  {2, ckReconnect,              MADB_OPT_FLAG_AUTO_RECONNECT},
-  {2, ckConnectPrompt,          MADB_OPT_FLAG_NO_PROMPT},
-  {2, ckCompressed,             MADB_OPT_FLAG_COMPRESSED_PROTO},
-  {3, ckIgnoreSchema,           MADB_OPT_FLAG_NO_SCHEMA},
-  {3, ckIgnoreSpace,            MADB_OPT_FLAG_IGNORE_SPACE},
-  {3, ckMultiStmt,              MADB_OPT_FLAG_MULTI_STATEMENTS},
-  {4, ckIgnoreSchema,           MADB_OPT_FLAG_NO_SCHEMA},
-  {4, ckEnableDynamicCursor,    MADB_OPT_FLAG_DYNAMIC_CURSOR},
-  {4, ckDisableDriverCursor,    MADB_OPT_FLAG_NO_DEFAULT_CURSOR},
-  {4, ckDontCacheForwardCursor, MADB_OPT_FLAG_NO_CACHE},
-  {4, ckForwardCursorOnly,      MADB_OPT_FLAG_FORWARD_CURSOR},
-  {4, ckReturnMatchedRows,      MADB_OPT_FLAG_FOUND_ROWS},
-  {4, ckEnableSQLAutoIsNull,    MADB_OPT_FLAG_AUTO_IS_NULL},
-  {4, ckPadCharFullLength,      MADB_OPT_FLAG_PAD_SPACE},
-  {4, ckNullDate,               MADB_OPT_FLAG_ZERO_DATE_TO_MIN},
-  {4, ckDebug,                  MADB_OPT_FLAG_DEBUG},
-  {4, ckReturnMatchedRows,      MADB_OPT_FLAG_FOUND_ROWS},
+  {1, rbPipe,                          MADB_OPT_FLAG_NAMED_PIPE},
+  {2, ckReconnect,                     MADB_OPT_FLAG_AUTO_RECONNECT},
+  {2, ckConnectPrompt,                 MADB_OPT_FLAG_NO_PROMPT},
+  {2, ckCompressed,                    MADB_OPT_FLAG_COMPRESSED_PROTO},
+  {2, ckUseMycnf,                      MADB_OPT_FLAG_USE_CNF},
+  {3, ckIgnoreSchema,                  MADB_OPT_FLAG_NO_SCHEMA},
+  {3, ckIgnoreSpace,                   MADB_OPT_FLAG_IGNORE_SPACE},
+  {3, ckMultiStmt,                     MADB_OPT_FLAG_MULTI_STATEMENTS},
+  {LASTPAGE, ckEnableDynamicCursor,    MADB_OPT_FLAG_DYNAMIC_CURSOR},
+  {LASTPAGE, ckDisableDriverCursor,    MADB_OPT_FLAG_NO_DEFAULT_CURSOR},
+  {LASTPAGE, ckDontCacheForwardCursor, MADB_OPT_FLAG_NO_CACHE},
+  {LASTPAGE, ckForwardCursorOnly,      MADB_OPT_FLAG_FORWARD_CURSOR},
+  {LASTPAGE, ckDontCacheForwardCursor, MADB_OPT_FLAG_NO_CACHE},
+  {LASTPAGE, ckReturnMatchedRows,      MADB_OPT_FLAG_FOUND_ROWS},
+  {LASTPAGE, ckEnableSQLAutoIsNull,    MADB_OPT_FLAG_AUTO_IS_NULL},
+  {LASTPAGE, ckPadCharFullLength,      MADB_OPT_FLAG_PAD_SPACE},
+  {LASTPAGE, ckNullDate,               MADB_OPT_FLAG_ZERO_DATE_TO_MIN},
+  {LASTPAGE, ckDebug,                  MADB_OPT_FLAG_DEBUG},
+  {LASTPAGE, ckReturnMatchedRows,      MADB_OPT_FLAG_FOUND_ROWS},
+  {LASTPAGE, ckIgnoreSpace,            MADB_OPT_FLAG_IGNORE_SPACE},
   /* last element */
   {0, 0, 0}
 };
@@ -102,11 +140,27 @@ MADB_OptionsMap OptionsMap[]= {
 void GetDialogFields(void);
 HBRUSH hbrBg= NULL;
 
+
+void DsnApplyDefaults(MADB_Dsn* Dsn)
+{
+  /* Setting default port number if Tcp/Ip selected, and the port is 0 */
+  if (Dsn->IsTcpIp && Dsn->Port == 0)
+  {
+    Dsn->Port= 3306;
+  }
+}
+
+
 my_bool SetDialogFields()
 {
   int  i= 0;
   MADB_Dsn *Dsn= (MADB_Dsn *)GetWindowLongPtr(GetParent(hwndTab[0]), DWLP_USER);
 
+  /* Basically - if dialog does not exist yet */
+  if (Dsn == NULL)
+  {
+    return TRUE;
+  }
   while (DsnMap[i].Key)
   {
     switch (DsnMap[i].Key->Type) {
@@ -130,6 +184,10 @@ my_bool SetDialogFields()
         SendDlgItemMessage(hwndTab[DsnMap[i].Page], DsnMap[i].Item, BM_SETCHECK,
                            Val ? BST_CHECKED : BST_UNCHECKED, 0);
       }
+      break;
+    case DSN_TYPE_CBOXGROUP:
+      SendDlgItemMessage(hwndTab[DsnMap[i].Page], DsnMap[i].Item, BM_SETCHECK,
+        (*GET_FIELD_PTR(Dsn, DsnMap[i].Key, char) & CBGROUP_BIT(i)) != '\0' ? BST_CHECKED : BST_UNCHECKED, 0);
     }
     i++;
   }
@@ -142,6 +200,8 @@ my_bool SetDialogFields()
     if (Val && OptionsMap[i].Item == rbPipe)
     {
       SendMessage(GetDlgItem(hwndTab[OptionsMap[i].Page], lblServerName), WM_SETTEXT, 0, (LPARAM)"Named pipe:");
+      ShowWindow(GetDlgItem(hwndTab[OptionsMap[i].Page], txtServerName), SW_HIDE);
+      ShowWindow(GetDlgItem(hwndTab[OptionsMap[i].Page], txtSocket), SW_SHOW);
 		  ShowWindow(GetDlgItem(hwndTab[OptionsMap[i].Page], lblPort), SW_HIDE);
       ShowWindow(GetDlgItem(hwndTab[OptionsMap[i].Page], txtPort), SW_HIDE);
     }
@@ -154,7 +214,7 @@ my_bool SetDialogFields()
 char *GetFieldStrVal(int Dialog, int Field, char* (*allocator)(size_t))
 {
   int rc;
-  size_t len= Edit_GetTextLength(GetDlgItem(hwndTab[Dialog], Field));
+  int len= Edit_GetTextLength(GetDlgItem(hwndTab[Dialog], Field));
   char *p;
 
   if (allocator)
@@ -167,7 +227,7 @@ char *GetFieldStrVal(int Dialog, int Field, char* (*allocator)(size_t))
   }
 
   if (p)
-    rc= Edit_GetText(GetDlgItem(hwndTab[Dialog], Field), p, len+1);
+    rc= Edit_GetText(GetDlgItem(hwndTab[Dialog], Field), p, len + 1);
   return p;      
 }
 
@@ -188,9 +248,11 @@ my_bool GetButtonState(int Dialog, int Button)
   return (my_bool)IsDlgButtonChecked(hwndTab[Dialog], Button);
 }
 
+
 my_bool SaveDSN(HWND hDlg, MADB_Dsn *Dsn)
 {
-  if (MADB_SaveDSN(Dsn))
+  DsnApplyDefaults(Dsn);
+  if (Dsn->SaveFile != NULL || MADB_SaveDSN(Dsn))
     return TRUE;
   MessageBox(hDlg, Dsn->ErrorMsg, "Error", MB_OK);
   return FALSE;
@@ -315,7 +377,17 @@ void GetDialogFields()
        *(int *)((char *)Dsn + DsnMap[i].Key->DsnOffset)= GetFieldIntVal(DsnMap[i].Page, DsnMap[i].Item);
        break;
     case DSN_TYPE_BOOL:
-      *(my_bool *)((char *)Dsn + DsnMap[i].Key->DsnOffset)=  GetButtonState(DsnMap[i].Page, DsnMap[i].Item);
+      *GET_FIELD_PTR(Dsn, DsnMap[i].Key, my_bool)=  IS_CB_CHECKED(i);
+      break;
+    case DSN_TYPE_CBOXGROUP:
+      if (IS_CB_CHECKED(i) != '\0')
+      {
+        CBGROUP_SETBIT(Dsn, i);
+      }
+      else
+      {
+        CBGROUP_RESETBIT(Dsn, i);
+      }
     }
     ++i;
   }
@@ -335,6 +407,7 @@ void DSN_Set_Database(SQLHANDLE Connection)
   SQLRETURN ret= SQL_ERROR;
   char Database[65];
   MADB_Dsn *Dsn= (MADB_Dsn *)GetWindowLongPtr(GetParent(hwndTab[0]), DWLP_USER);
+  HWND DbCombobox= GetDlgItem(hwndTab[1], cbDatabase);
   
 
   if (DBFilled)
@@ -349,15 +422,17 @@ void DSN_Set_Database(SQLHANDLE Connection)
     goto end;
 
   SQLBindCol(Stmt, 1, SQL_C_CHAR, Database, 65, 0);
-  ComboBox_ResetContent(GetDlgItem(hwndTab[1], cbDatabase));
+  ComboBox_ResetContent(DbCombobox);
   while (SQLFetch(Stmt) == SQL_SUCCESS)
-    ComboBox_InsertString(GetDlgItem(hwndTab[1], cbDatabase), -1, Database);
+  {
+    ComboBox_InsertString(DbCombobox, -1, Database);
+  }
   if (Dsn->Catalog)
   {
-    int Idx= ComboBox_FindString(GetDlgItem(hwndTab[2], cbDatabase), 0, Dsn->Catalog);
-    ComboBox_SetCurSel(GetDlgItem(hwndTab[2], cbDatabase), Idx);
+    int Idx= ComboBox_FindString(DbCombobox, 0, Dsn->Catalog);
+    ComboBox_SetCurSel(DbCombobox, Idx);
   }
-  ComboBox_SetMinVisible(GetDlgItem(hwndTab[1], cbDatabase),5);
+  ComboBox_SetMinVisible(GetDlgItem(hwndTab[1], cbDatabase), 5);
   DBFilled= TRUE;
 
 end:
@@ -407,56 +482,236 @@ end:
 	  SQLFreeHandle(SQL_HANDLE_STMT, (SQLHANDLE)Stmt);
 }
 
-void MADB_WIN_TestDsn(my_bool ShowSuccess)
+
+static SQLRETURN TestDSN(MADB_Dsn *Dsn, SQLHANDLE *Conn, SQLCHAR *ConnStrBuffer)
 {
-  MYSQL *mysql= mysql_init(NULL);
   SQLHANDLE Connection= NULL;
-  SQLRETURN ret;
-  MADB_Dsn *Dsn= (MADB_Dsn *)GetWindowLongPtr(GetParent(hwndTab[0]), DWLP_USER);
-  
-  GetDialogFields();
+  SQLRETURN Result;
+  SQLCHAR LocalBuffer[CONNSTR_BUFFER_SIZE], *ConnStr= LocalBuffer;
+  char *InitCommand= Dsn->InitCommand;
+  char *DsName=      Dsn->DSNName;
+  char *Description= Dsn->Description;
+
+  Dsn->InitCommand= NULL;
+  Dsn->DSNName=     NULL;
+  Dsn->Description= NULL;
+
+  if (ConnStrBuffer != NULL)
+  {
+    ConnStr= ConnStrBuffer;
+  }
+
+  DsnApplyDefaults(Dsn);
+  /* If defaults has changed actual values - let them be reflected in the dialog(if it exists - SetDialogFields
+     cares about that) */
+  SetDialogFields();
+  MADB_DsnToString(Dsn, ConnStr, CONNSTR_BUFFER_SIZE);
 
   SQLAllocHandle(SQL_HANDLE_DBC, Environment, (SQLHANDLE *)&Connection);
   assert(Connection != NULL);
-  
-  ret= ((MADB_Dbc *)Connection)->Methods->ConnectDB((MADB_Dbc *)Connection, Dsn);
+  Result= SQLDriverConnect(Connection, NULL, ConnStr, CONNSTR_BUFFER_SIZE, NULL, 0, NULL, SQL_DRIVER_NOPROMPT);
+
+  Dsn->InitCommand= InitCommand;
+  Dsn->DSNName= DsName;
+  Dsn->Description= Description;
+
+  if (Conn != NULL)
+  {
+    *Conn= Connection;
+  }
+  else
+  {
+    SQLDisconnect(Connection);
+    SQLFreeHandle(SQL_HANDLE_DBC, Connection);
+  }
+
+  return Result;
+}
+
+
+/* Connstr has to be null-terminated */
+char* HidePwd(char *ConnStr)
+{
+  char *Ptr= ConnStr;
+
+  while (*Ptr)
+  {
+    BOOL IsPwd=          FALSE;
+    char *KeyValBorder= strchr(Ptr, '=');
+    char StopChr=        ';';
+
+    Ptr= ltrim(Ptr);
+
+    if (_strnicmp(Ptr, "PWD", 3) == 0 || _strnicmp(Ptr, "PASSWORD", 8) == 0)
+    {
+      IsPwd= TRUE;
+    }
+    if (KeyValBorder != NULL)
+    {
+      Ptr= ltrim(KeyValBorder + 1);
+    }
+    if (*Ptr == '{')
+    {
+      StopChr= '}';
+    }
+    while (*Ptr && *Ptr != StopChr)
+    {
+      if (IsPwd)
+      {
+        *Ptr= '*';
+      }
+      ++Ptr;
+    }
+    /* Do not move if we already at the terminating null */
+    if (*Ptr)
+    {
+      ++Ptr;
+    }
+  }
+
+  return ConnStr;
+}
+
+
+void MADB_WIN_TestDsn(my_bool ShowSuccess)
+{
+  SQLHANDLE Connection;
+  SQLRETURN ret;
+  MADB_Dsn *Dsn= (MADB_Dsn *)GetWindowLongPtr(GetParent(hwndTab[0]), DWLP_USER);
+  SQLCHAR ConnStr[CONNSTR_BUFFER_SIZE];
+
+  GetDialogFields();
+
+  ret= TestDSN(Dsn, &Connection, ConnStr);
 
   if (ShowSuccess)
   {
     char Info[1024];
-	if (ret == SQL_SUCCESS)
-      my_snprintf(Info, 1024, "Connection successfully established\n\nServer information: %s", mysql_get_server_info(((MADB_Dbc *)Connection)->mariadb));
-    MessageBox(hwndTab[CurrentPage],  (ret == SQL_SUCCESS) ? Info : 
-	         ((MADB_Dbc *)Connection)->Error.SqlErrorMsg, "Connection test",MB_ICONINFORMATION| MB_OK);
+
+    HidePwd(ConnStr);
+
+    if (SQL_SUCCEEDED(ret))
+    {
+      char DbmsName[16], DbmsVer[16];
+      SQLGetInfo(Connection, SQL_DBMS_NAME, DbmsName, sizeof(DbmsName), NULL);
+      SQLGetInfo(Connection, SQL_DBMS_VER, DbmsVer, sizeof(DbmsVer), NULL);
+      _snprintf(Info, sizeof(Info), "Connection successfully established\n\nServer Information: %s %s\n\nConnection String:\n\n%s", DbmsName, DbmsVer, ConnStr);
+      MessageBox(hwndTab[CurrentPage], Info, "Connection test", MB_ICONINFORMATION| MB_OK);
+    }
+    else
+    {
+      SQLCHAR SqlState[6], ErrMsg[SQL_MAX_MESSAGE_LENGTH];
+      SQLGetDiagRec(SQL_HANDLE_DBC,   Connection, 1, SqlState, NULL, ErrMsg, SQL_MAX_MESSAGE_LENGTH, NULL);
+      _snprintf(Info, sizeof(Info), "Connection failed: [%s] %s\n\nConnection String:\n\n%s", SqlState, ErrMsg, ConnStr);
+      MessageBox(hwndTab[CurrentPage], Info, "Connection test", MB_ICONINFORMATION| MB_OK);
+    }
   }
 
-  if (ret == SQL_SUCCESS)
+  if (SQL_SUCCEEDED(ret))
   {
     ConnectionOK= TRUE;
     DSN_Set_CharacterSets(Connection);
     DSN_Set_Database(Connection);
+
+    SQLDisconnect(Connection);
   }
 
-  SQLDisconnect(Connection);
   SQLFreeHandle(SQL_HANDLE_DBC, Connection);
 }
 
 
+INT_PTR SelectPath(HWND ParentWnd, int BoundEditId, const wchar_t *Caption, BOOL FolderPath, BOOL OpenCurrentSelection)
+{
+  if (SUCCEEDED(CoInitializeEx(NULL, COINIT_APARTMENTTHREADED | COINIT_DISABLE_OLE1DDE)))
+  {
+    IFileDialog *PathDialog;
+
+    if (SUCCEEDED(CoCreateInstance(&CLSID_FileOpenDialog, NULL, CLSCTX_ALL, 
+                                   &IID_IFileOpenDialog, (void**)&PathDialog)))
+    {
+      HWND        BoundEditWnd= GetDlgItem(ParentWnd, BoundEditId);
+      DWORD       DialogOptions;
+      TCHAR       Path[MAX_PATH];
+      IShellItem *SelectedItem;
+      int         Length;
+      LPWSTR      wPath;
+
+      if (FolderPath && SUCCEEDED(PathDialog->lpVtbl->GetOptions(PathDialog, &DialogOptions)))
+      {
+        PathDialog->lpVtbl->SetOptions(PathDialog, DialogOptions | FOS_PICKFOLDERS);
+      }
+
+      Edit_GetText(BoundEditWnd, Path, sizeof(Path));
+      Length= MultiByteToWideChar(GetACP(), 0, Path, -1, NULL, 0);
+
+      wPath= (SQLWCHAR *)malloc(Length * sizeof(SQLWCHAR));
+      if (wPath == NULL)
+      {
+        return FALSE;
+      }
+
+      MultiByteToWideChar(GetACP(), 0, Path, -1, wPath, Length);
+
+      SHCreateItemFromParsingName(wPath, NULL, &IID_IShellItem, (void**)&SelectedItem);
+      if (OpenCurrentSelection)
+      {
+        PathDialog->lpVtbl->SetFolder(PathDialog, SelectedItem);
+      }
+      else
+      {
+        PathDialog->lpVtbl->SetDefaultFolder(PathDialog, SelectedItem);
+      }
+      SelectedItem->lpVtbl->Release(SelectedItem);
+
+      free(wPath);
+
+      PathDialog->lpVtbl->SetTitle(PathDialog, Caption);
+
+      if (SUCCEEDED(PathDialog->lpVtbl->Show(PathDialog, ParentWnd)))
+      {
+        if (SUCCEEDED(PathDialog->lpVtbl->GetResult(PathDialog, &SelectedItem)))
+        {
+          LPWSTR SelectedValue;
+          if (SUCCEEDED(SelectedItem->lpVtbl->GetDisplayName(SelectedItem, SIGDN_FILESYSPATH, &SelectedValue)))
+          {
+            BOOL Error;
+
+            /* TODO: I guess conversions from/to utf8 has to be done when syncing edits with DSN */
+            WideCharToMultiByte(GetACP(), 0, SelectedValue, -1, Path, sizeof(Path), NULL, &Error);
+            Edit_SetText(BoundEditWnd, Path);
+            CoTaskMemFree(SelectedValue);
+
+            return TRUE;
+          }
+          SelectedItem->lpVtbl->Release(SelectedItem);
+        }
+      }
+      PathDialog->lpVtbl->Release(PathDialog);
+    }
+    CoUninitialize();
+  }
+
+  return FALSE;
+}
+
+
 INT_PTR CALLBACK DialogDSNProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
 {
-	  switch(uMsg)
+  INT_PTR res;
+
+	switch(uMsg)
   {
   case WM_CTLCOLORDLG:
     if (!hbrBg)
       hbrBg= CreateSolidBrush(RGB(255,255,255));
-      return (LONG)hbrBg;
+      return (INT_PTR)hbrBg;
     break;
    case WM_CTLCOLORSTATIC:
    {
      HDC hdcStatic = (HDC)wParam;
      SetTextColor(hdcStatic, RGB(0, 0, 0));
      SetBkMode(hdcStatic, TRANSPARENT);
-     return (LONG)hbrBg;
+     return (INT_PTR)hbrBg;
   }
   case WM_COMMAND:
     switch(LOWORD(wParam))
@@ -483,21 +738,54 @@ INT_PTR CALLBACK DialogDSNProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lPara
                  "Warning", MB_OK);
         return FALSE;
       }
-	    return TRUE; 
-  case rbTCP:
-	case rbPipe:
-		if (HIWORD(wParam) == BN_CLICKED)
-		{
-	    SendMessage(GetDlgItem(hwndTab[CurrentPage], lblServerName), WM_SETTEXT, 0,
-			 (LOWORD(wParam) == rbTCP) ? (LPARAM)"Server name:" : (LPARAM)"Named pipe:");
-		  ShowWindow(GetDlgItem(hwndTab[CurrentPage], lblPort),
-			  (LOWORD(wParam) == rbTCP) ? SW_SHOW : SW_HIDE);
-		  ShowWindow(GetDlgItem(hwndTab[CurrentPage], txtPort),
-			  (LOWORD(wParam) == rbTCP) ? SW_SHOW : SW_HIDE);
-		}
-		return TRUE;
-
-    }
+	    return TRUE;
+    case pbPlugindirBrowse:
+      return SelectPath(hDlg, txtPluginDir,   L"Select Plugins Directory",          /* Folder*/ TRUE, TRUE);
+    case pbCaPathBrowse:
+      res=   SelectPath(hDlg, txtSslCaPath,   L"Select CA Path",                                TRUE, OpenCurSelection);
+      OpenCurSelection= OpenCurSelection && !res;
+      return res;
+    case pbKeyBrowse:
+      res=   SelectPath(hDlg, txtSslKey,      L"Select Client Private Key File",    /* File */  FALSE, OpenCurSelection);
+      OpenCurSelection= OpenCurSelection && !res;
+      return res;
+    case pbCertBrowse:
+      res=   SelectPath(hDlg, txtSslCert,     L"Select Public Key Certificate File",            FALSE, OpenCurSelection);
+      OpenCurSelection= OpenCurSelection && !res;
+      return res;
+    case pbCaCertBrowse:
+      res=   SelectPath(hDlg, txtSslCertAuth, L"Select Certificate Authority Certificate File", FALSE,OpenCurSelection);
+      OpenCurSelection= OpenCurSelection && !res;
+      return res;
+    case pbServerKeyBrowse:
+      res=   SelectPath(hDlg, txtServerKey, L"Select Server Public Key File", FALSE, OpenCurSelection);
+      OpenCurSelection= OpenCurSelection && !res;
+      return res;
+    case pbFpListBrowse:
+      res=   SelectPath(hDlg, txtTlsPeerFpList, L"Select File with SHA1 fingerprints of server certificates", FALSE, OpenCurSelection);
+      OpenCurSelection= OpenCurSelection && !res;
+      return res;
+    case pbCrlBrowse:
+      res = SelectPath(hDlg, txtCrl, L"Select PEM File Certificate Revocation List(CRL)", FALSE, OpenCurSelection);
+      OpenCurSelection = OpenCurSelection && !res;
+      return res;
+    case rbTCP:
+	  case rbPipe:
+		  if (HIWORD(wParam) == BN_CLICKED)
+		  {
+	      SendMessage(GetDlgItem(hwndTab[CurrentPage], lblServerName), WM_SETTEXT, 0,
+			   (LOWORD(wParam) == rbTCP) ? (LPARAM)"Server name:" : (LPARAM)"Named pipe:");
+		    ShowWindow(GetDlgItem(hwndTab[CurrentPage], lblPort),
+			    (LOWORD(wParam) == rbTCP) ? SW_SHOW : SW_HIDE);
+		    ShowWindow(GetDlgItem(hwndTab[CurrentPage], txtPort),
+			    (LOWORD(wParam) == rbTCP) ? SW_SHOW : SW_HIDE);
+        ShowWindow(GetDlgItem(hwndTab[CurrentPage], txtServerName),
+          (LOWORD(wParam) == rbTCP) ? SW_SHOW : SW_HIDE);
+        ShowWindow(GetDlgItem(hwndTab[CurrentPage], txtSocket),
+          (LOWORD(wParam) == rbTCP) ? SW_HIDE : SW_SHOW);
+		  }
+		  return TRUE;
+      }
     break;
 	}
 	return FALSE;
@@ -510,7 +798,7 @@ INT_PTR CALLBACK DialogProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
   case WM_CTLCOLORDLG:
     if (!hbrBg)
       hbrBg= CreateSolidBrush(RGB(255,255,255));
-      return (LONG)hbrBg;
+      return (INT_PTR)hbrBg;
     break;
   case WM_CLOSE:
     if(MessageBox(hDlg, TEXT("Close the program?"), TEXT("Close"),
@@ -526,7 +814,7 @@ INT_PTR CALLBACK DialogProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
     return TRUE;
   case WM_INITDIALOG:
   {
-	  static int Dialogs[] = {Page_0, Page_1, Page_2, Page_3, Page_4};
+	  static int Dialogs[] = {Page_0, Page_1, Page_2, Page_3, Page_4, Page_5};
   	int i;
   	RECT rc;
   	GetWindowRect(hDlg, &rc);
@@ -551,6 +839,7 @@ INT_PTR CALLBACK DialogProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
   return FALSE;
 }
 
+
 BOOL APIENTRY DllMain( HMODULE hModule,
                        DWORD  ul_reason_for_call,
                        LPVOID lpReserved
@@ -560,6 +849,7 @@ BOOL APIENTRY DllMain( HMODULE hModule,
   return TRUE;
 }
 
+
 void CenterWindow(HWND hwndWindow)
 {
   RECT rectWindow;
@@ -578,30 +868,29 @@ void CenterWindow(HWND hwndWindow)
   
 }
 
-my_bool DSNDialog(HWND hwndParent,
-                  WORD fRequest,
-                  LPCSTR lpszDriver,
-                  LPCSTR lpszAttributes,
-                  MADB_Dsn *Dsn)
+
+BOOL DSNDialog(HWND     hwndParent,
+               WORD     fRequest,
+               LPCSTR   lpszDriver,
+               LPCSTR   lpszAttributes,
+               MADB_Dsn *Dsn)
 {
-  MSG msg;
-  BOOL ret;
-  char *DsnName= NULL;
+  MSG     msg;
+  BOOL    ret;
+  char    *DsnName=  NULL;
   my_bool DsnExists= FALSE;
+  char    Delimiter= ';';
 
   if (Dsn->isPrompt < 0 || Dsn->isPrompt > MAODBC_PROMPT_REQUIRED)
   {
     Dsn->isPrompt= MAODBC_CONFIG;
   }
 
-  EffectiveDisabledPages=     DisabledPages[Dsn->isPrompt];
-  EffectiveDisabledControls=  DisabledControls[Dsn->isPrompt];
-
   if (lpszAttributes)
+  {
+    Delimiter= '\0';
     DsnName= strchr((char *)lpszAttributes, '=');
-  
-  if (lpszDriver)
-    MADB_DriverGet((char *)lpszDriver);
+  }
 
   if (DsnName)
   {
@@ -610,7 +899,9 @@ my_bool DSNDialog(HWND hwndParent,
     if (!Dsn->isPrompt && !SQLValidDSN(DsnName))
     {
       if (hwndParent)
+      {
         MessageBox(hwndParent, "Validation of data source name failed", "Error", MB_ICONERROR | MB_OK);
+      }
       return FALSE;
     }
   }
@@ -621,60 +912,106 @@ my_bool DSNDialog(HWND hwndParent,
   }
   else if (DsnName && Dsn)
   {
-    /* Need to free current value in Dsn->DSNName */
-    MADB_SUBSTITUTE(Dsn->DSNName, _strdup(DsnName));
+    MADB_RESET(Dsn->DSNName, DsnName);
   }
 
   /* Even if DsnName invalid(in case of prompt) - we should not have problem */
   DsnExists= MADB_DSN_Exists(DsnName);
- 
-  InitCommonControls();
-
-  if (lpszDriver)
-    Dsn->Driver= _strdup(lpszDriver);
 
   SQLAllocHandle(SQL_HANDLE_ENV, SQL_NULL_HANDLE, &Environment);
   SQLSetEnvAttr(Environment, SQL_ATTR_ODBC_VERSION, (SQLPOINTER)SQL_OV_ODBC3, 0);
 
-  if (fRequest == ODBC_ADD_DSN)
-  {
-    if (DsnExists && hwndParent)
-    {
-      if (MessageBox(hwndParent, "Data source name already exists, do you want to replace it?", 
-                      "Question", MB_ICONQUESTION | MB_YESNO) != IDYES)
-        return FALSE;
-    }
-    Dsn->IsTcpIp= 1;
-  }
-  else
+  switch (fRequest)
   {
+    case ODBC_ADD_DSN:
+      if (DsnExists)
+      {
+        if (hwndParent)
+        {
+          if (MessageBox(hwndParent, "Data source name already exists, do you want to replace it?",
+            "Question", MB_ICONQUESTION | MB_YESNO) != IDYES)
+          {
+            return FALSE;
+          }
+        }
+      }
+      
+      MADB_DSN_SetDefaults(Dsn);
+      MADB_ParseConnString(Dsn, (char *)lpszAttributes, SQL_NTS, Delimiter);
+
+      /* Need to set driver after connstring parsing, and before saving */
+      if (lpszDriver)
+      {
+        Dsn->Driver= _strdup(lpszDriver);
+      }
+
+      /* If we don't have parent window - we are not supposed to show the dialog, but only
+         perform operation, if sufficient data has been provided */
+      if (hwndParent == NULL)
+      {
+        if (SQL_SUCCEEDED(TestDSN(Dsn, NULL, NULL)))
+        {
+          return MADB_SaveDSN(Dsn);
+        }
+        else
+        {
+          return FALSE;
+        }
+      }
+
+      break;
+
+    case ODBC_CONFIG_DSN:
+
     /* i.e. not a prompt */
-    if (Dsn->isPrompt == MAODBC_CONFIG)
+    if (Dsn->isPrompt == MAODBC_CONFIG && Dsn->SaveFile == NULL)
     {
       if (!DsnExists)
       {
-        MessageBox(0, "Data source name not found", "Error", MB_ICONERROR | MB_OK);
+        if (hwndParent != NULL)
+        {
+          MessageBox(0, "Data source name not found", "Error", MB_ICONERROR | MB_OK);
+        }
         return FALSE;
       }
-      else if (!MADB_ReadDSN(Dsn, (char *)lpszAttributes, TRUE))
+      else if (!MADB_ReadConnString(Dsn, (char *)lpszAttributes, SQL_NTS, Delimiter))
       {
         SQLPostInstallerError(ODBC_ERROR_INVALID_DSN, Dsn->ErrorMsg);
         return FALSE;
       }
+
+      if (hwndParent == NULL)
+      {
+        if (SQL_SUCCEEDED(TestDSN(Dsn, NULL, NULL)))
+        {
+          return MADB_SaveDSN(Dsn);
+        }
+        else
+        {
+          return FALSE;
+        }
+      }
     }
   }
 
+  InitCommonControls();
+
+  EffectiveDisabledPages=     DisabledPages[Dsn->isPrompt];
+  EffectiveDisabledControls=  DisabledControls[Dsn->isPrompt];
+
+  if (fRequest == ODBC_ADD_DSN && Dsn->isPrompt == MAODBC_CONFIG && Dsn->DSNName != NULL)
+  {
+    EffectiveDisabledControls= PromptDisabledControls;
+  }
+
   notCanceled= TRUE;
   hwndMain= CreateDialogParam(hInstance, MAKEINTRESOURCE(IDD_DIALOG1), 0, DialogProc, 0);
-  SetWindowLongPtr(hwndMain, DWLP_USER, (LONG)Dsn);
+  SetWindowLongPtr(hwndMain, DWLP_USER, (LONG_PTR)Dsn);
 
   /* Setting first not disabled page */
   CurrentPage= -1;
   SetPage(hwndMain, 1);
 
-  Edit_SetReadOnly(GetDlgItem(hwndTab[0], txtDsnName), 
-                   (hwndParent && DsnName && fRequest == ODBC_ADD_DSN) ? TRUE : FALSE);
-
   SetDialogFields();
   CenterWindow(hwndMain);
   ShowWindow(hwndMain, SW_SHOW);
@@ -691,9 +1028,15 @@ my_bool DSNDialog(HWND hwndParent,
 
   SQLFreeHandle(SQL_HANDLE_ENV, Environment);
 
+  if (notCanceled && fRequest == ODBC_CONFIG_DSN && DsnName != NULL && strcmp(DsnName, Dsn->DSNName) != 0)
+  {
+    SQLRemoveDSNFromIni(DsnName);
+  }
+
   return notCanceled;
 }
 
+
 BOOL INSTAPI ConfigDSN(
      HWND     hwndParent,
      WORD     fRequest,
