@@ -1,6 +1,6 @@
 /************************************************************************************
-   Copyright (C) 2013, 2014 SkySQL AB
-   
+   Copyright (C) 2013, 2019 MariaDB Corporation AB
+
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
@@ -18,6 +18,8 @@
 *************************************************************************************/
 #include <ma_odbc.h>
 
+extern Client_Charset utf8;
+
 /* To keep the source code smaller we use the following structure to check if a field
    identifier is valid for a given descriptor type */
 struct st_ma_desc_fldid MADB_DESC_FLDID[]=
@@ -51,7 +53,7 @@ struct st_ma_desc_fldid MADB_DESC_FLDID[]=
   {SQL_DESC_NUM_PREC_RADIX, {MADB_DESC_RW, MADB_DESC_RW, MADB_DESC_RW, MADB_DESC_READ}},
   {SQL_DESC_OCTET_LENGTH, {MADB_DESC_RW, MADB_DESC_RW, MADB_DESC_RW, MADB_DESC_READ}},
   {SQL_DESC_OCTET_LENGTH_PTR,{MADB_DESC_RW, MADB_DESC_RW, MADB_DESC_NONE, MADB_DESC_NONE}},
-  {SQL_DESC_PARAMETER_TYPE,{MADB_DESC_NONE, MADB_DESC_NONE, MADB_DESC_RW, MADB_DESC_RW}},
+  {SQL_DESC_PARAMETER_TYPE,{MADB_DESC_NONE, MADB_DESC_NONE, MADB_DESC_RW, MADB_DESC_NONE}},
   {SQL_DESC_PRECISION,{MADB_DESC_RW, MADB_DESC_RW, MADB_DESC_RW, MADB_DESC_READ}},
 #if (ODBCVER >= 0x0350)
   {SQL_DESC_ROWVER,{MADB_DESC_NONE, MADB_DESC_NONE, MADB_DESC_READ, MADB_DESC_READ}},
@@ -69,7 +71,7 @@ struct st_ma_desc_fldid MADB_DESC_FLDID[]=
 };
 
 /* {{{ MADB_DescInit */
-MADB_Desc *MADB_DescInit(enum enum_madb_desc_type DescType, my_bool isExternal)
+MADB_Desc *MADB_DescInit(MADB_Dbc *Dbc,enum enum_madb_desc_type DescType, my_bool isExternal)
 {
   MADB_Desc *Desc;
   
@@ -77,20 +79,29 @@ MADB_Desc *MADB_DescInit(enum enum_madb_desc_type DescType, my_bool isExternal)
     return NULL;
 
   Desc->DescType= DescType;
+  MADB_PutErrorPrefix(Dbc, &Desc->Error);
 
-  if (my_init_dynamic_array(&Desc->Records, sizeof(MADB_DescRecord), 0, 0))
+  if (MADB_InitDynamicArray(&Desc->Records, sizeof(MADB_DescRecord), 0, MADB_DESC_INIT_REC_NUM))
   {
     MADB_FREE(Desc);
-    Desc= NULL;
+    return NULL;
   }
-  if (isExternal && my_init_dynamic_array(&Desc->Stmts, sizeof(MADB_Stmt**), 0, 0))
+  if (isExternal)
   {
-    MADB_DescFree(Desc, FALSE);
-    Desc= NULL;
+    if (MADB_InitDynamicArray(&Desc->Stmts, sizeof(MADB_Stmt**), 0, MADB_DESC_INIT_STMT_NUM))
+    {
+      MADB_DescFree(Desc, FALSE);
+      return NULL;
+    }
+    else
+    {
+      Desc->Dbc= Dbc;
+      /* MADB_DescInit call for explicit descriptor is in critical section */
+      Desc->ListItem.data= (void *)Desc;
+      Dbc->Descrs= MADB_ListAdd(Dbc->Descrs, &Desc->ListItem);
+    }
   }
-  if (Desc)
-    Desc->AppType= isExternal;
-
+  Desc->AppType= isExternal;
   Desc->Header.ArraySize= 1;
  
   return Desc;
@@ -113,7 +124,7 @@ SQLRETURN MADB_DescFree(MADB_Desc *Desc, my_bool RecordsOnly)
     MADB_FREE(Record->InternalBuffer);
     MADB_FREE(Record->DefaultValue);
  
-   if (Desc->DescType == MADB_DESC_IRD)
+    if (Desc->DescType == MADB_DESC_IRD)
     {
       MADB_FREE(Record->CatalogName);
       MADB_FREE(Record->BaseCatalogName);
@@ -122,161 +133,455 @@ SQLRETURN MADB_DescFree(MADB_Desc *Desc, my_bool RecordsOnly)
       MADB_FREE(Record->ColumnName);
       MADB_FREE(Record->TableName);
       MADB_FREE(Record->TypeName);
+    }
+    else if(Desc->DescType == MADB_DESC_IPD)
+    {
+      MADB_FREE(Record->TypeName);
+    }
+  }
+  MADB_DeleteDynamic(&Desc->Records);
 
-    }
-  }
-  delete_dynamic(&Desc->Records);
-  for (i=0; i < Desc->Stmts.elements; i++)
+  Desc->Header.Count= 0;
+  if (Desc->AppType)
   {
-    MADB_Stmt **XStmt= ((MADB_Stmt **)Desc->Stmts.buffer) + i;
-    MADB_Stmt *Stmt= *XStmt;
-    switch(Desc->DescType) {
-    case MADB_DESC_ARD:
-      Stmt->Ard=Stmt->IArd;
-      break;
-    case MADB_DESC_APD:
-      Stmt->Apd= Stmt->IApd;
-      break;
+    EnterCriticalSection(&Desc->Dbc->ListsCs);
+    for (i=0; i < Desc->Stmts.elements; i++)
+    {
+      MADB_Stmt **XStmt= ((MADB_Stmt **)Desc->Stmts.buffer) + i;
+      MADB_Stmt *Stmt= *XStmt;
+      switch(Desc->DescType) {
+      case MADB_DESC_ARD:
+        Stmt->Ard=Stmt->IArd;
+        break;
+      case MADB_DESC_APD:
+        Stmt->Apd= Stmt->IApd;
+        break;
+      }
     }
+    MADB_DeleteDynamic(&Desc->Stmts);
+  
+    Desc->Dbc->Descrs= MADB_ListDelete(Desc->Dbc->Descrs, &Desc->ListItem);
+    LeaveCriticalSection(&Desc->Dbc->ListsCs);
   }
-  delete_dynamic(&Desc->Stmts);
+  
   if (!RecordsOnly)
     MADB_FREE(Desc);
   return SQL_SUCCESS;
 }
 /* }}} */ 
 
+/* {{{ MADB_SetIrdRecord */
+my_bool
+MADB_SetIrdRecord(MADB_Stmt *Stmt, MADB_DescRecord *Record, MYSQL_FIELD *Field)
+{
+  MY_CHARSET_INFO cs;
+  MARIADB_CHARSET_INFO *FieldCs;
+
+  if (Record == NULL)
+  {
+    return 1;
+  }
+
+  mariadb_get_infov(Stmt->Connection->mariadb, MARIADB_CONNECTION_MARIADB_CHARSET_INFO, (void*)&cs);
+
+  MADB_RESET(Record->CatalogName,    Field->db);
+  MADB_RESET(Record->TableName,      Field->table);
+  MADB_RESET(Record->ColumnName,     Field->name);
+  MADB_RESET(Record->BaseTableName,  Field->org_table);
+  MADB_RESET(Record->BaseColumnName, Field->org_name);
+  Record->AutoUniqueValue= (Field->flags & AUTO_INCREMENT_FLAG) ? SQL_TRUE : SQL_FALSE;
+  Record->CaseSensitive=   (Field->flags & BINARY_FLAG) ? SQL_TRUE : SQL_FALSE;
+  Record->Nullable= ( (Field->flags & NOT_NULL_FLAG) &&
+                      !Record->AutoUniqueValue &&
+                      Field->type != MYSQL_TYPE_TIMESTAMP) ? SQL_NO_NULLS : SQL_NULLABLE;
+  Record->Unsigned= (Field->flags & UNSIGNED_FLAG) ? SQL_TRUE : SQL_FALSE;
+  /* We assume it might be updatable if tablename exists */
+  Record->Updateable= (Field->table && Field->table[0]) ? SQL_ATTR_READWRITE_UNKNOWN : SQL_ATTR_READONLY;
+
+  /*
+      RADIX:
+      If the data type in the SQL_DESC_TYPE field is an approximate numeric data type, this SQLINTEGER field 
+      contains a value of 2 because the SQL_DESC_PRECISION field contains the number of bits. 
+      If the data type in the SQL_DESC_TYPE field is an exact numeric data type, this field contains a value of 10 
+      because the SQL_DESC_PRECISION field contains the number of decimal digits. 
+      This field is set to 0 for all non-numeric data types.
+  */
+  switch (Field->type) {
+  case MYSQL_TYPE_DECIMAL:
+  case MYSQL_TYPE_NEWDECIMAL:
+    Record->NumPrecRadix= 10;
+    Record->Scale= Field->decimals;
+    Record->Precision= (SQLSMALLINT)Field->length - test(Record->Unsigned == SQL_FALSE) - test(Record->Scale > 0);
+    if (Record->Precision == 0)
+    {
+      Record->Precision= Record->Scale;
+    }
+    break;
+  case MYSQL_TYPE_FLOAT:
+    Record->NumPrecRadix= 2;
+    Record->Precision= (SQLSMALLINT)Field->length - 2;
+    //Record->Scale= Field->decimals;
+    break;
+  case MYSQL_TYPE_DOUBLE:
+  case MYSQL_TYPE_TINY:
+  case MYSQL_TYPE_SHORT:
+  case MYSQL_TYPE_INT24:
+  case MYSQL_TYPE_LONG:
+  case MYSQL_TYPE_LONGLONG:
+  case MYSQL_TYPE_YEAR:
+    Record->NumPrecRadix= 10;
+    break;
+  case MYSQL_TYPE_TIMESTAMP:
+  case MYSQL_TYPE_TIME:
+  case MYSQL_TYPE_DATETIME:
+    Record->Scale= Field->decimals;
+  default:
+    Record->NumPrecRadix= 0;
+    break;
+  }
+
+  Record->ConciseType= MapMariadDbToOdbcType(Field);
+
+  if (0)//Stmt->Connection->IsAnsi == '\0')
+  {
+    switch (Record->ConciseType) {
+    case SQL_CHAR:        Record->ConciseType= SQL_WCHAR; break;
+    case SQL_VARCHAR:     Record->ConciseType= SQL_WVARCHAR; break;
+    case SQL_LONGVARCHAR: Record->ConciseType= SQL_WLONGVARCHAR; break;
+    }
+  }
+  /* 
+      TYPE:
+      For the datetime and interval data types, this field returns the verbose data type: SQL_DATETIME or SQL_INTERVAL.
+  */
+  Record->Type= (Record->ConciseType ==  SQL_TYPE_DATE || Record->ConciseType ==  SQL_DATE ||
+                 Record->ConciseType ==  SQL_TYPE_TIME || Record->ConciseType ==  SQL_TIME ||
+                 Record->ConciseType ==  SQL_TYPE_TIMESTAMP || Record->ConciseType == SQL_TIMESTAMP) ?
+                    SQL_DATETIME : Record->ConciseType;
+
+  switch(Record->ConciseType) {
+  case SQL_TYPE_DATE:
+    Record->DateTimeIntervalCode= SQL_CODE_DATE;
+    break;
+  case SQL_TYPE_TIME:
+    Record->DateTimeIntervalCode= SQL_CODE_TIME;
+    break;
+  case SQL_TYPE_TIMESTAMP:
+    Record->DateTimeIntervalCode= SQL_CODE_TIMESTAMP;
+    break;
+  }
+  /* 
+      SEARCHABLE:
+      Columns of type SQL_LONGVARCHAR and SQL_LONGVARBINARY usually return SQL_PRED_CHAR.
+  */
+  Record->Searchable= (Record->ConciseType == SQL_LONGVARCHAR ||
+                       Record->ConciseType == SQL_WLONGVARCHAR ||
+                       Record->ConciseType == SQL_LONGVARBINARY) ? SQL_PRED_CHAR : SQL_SEARCHABLE;
+
+  Record->DisplaySize= MADB_GetDisplaySize(Field, mariadb_get_charset_by_nr(Field->charsetnr));
+  Record->OctetLength= MADB_GetOctetLength(Field, cs.mbmaxlen);
+  FieldCs= mariadb_get_charset_by_nr(Field->charsetnr);
+  Record->Length= MADB_GetDataSize(Record->ConciseType, Field->length, Record->Unsigned == SQL_TRUE,
+                                   Record->Precision, Record->Scale, FieldCs!= NULL ? FieldCs->char_maxlen : 1);
+    
+  MADB_RESET(Record->TypeName, MADB_GetTypeName(Field));
+
+  switch(Field->type) {
+  case MYSQL_TYPE_BLOB:
+  case MYSQL_TYPE_LONG_BLOB:
+  case MYSQL_TYPE_MEDIUM_BLOB:
+  case MYSQL_TYPE_STRING:
+  case MYSQL_TYPE_TINY_BLOB:
+  case MYSQL_TYPE_VAR_STRING:
+    if (Field->flags & BINARY_FLAG)
+    {
+      Record->LiteralPrefix= "0x";
+      Record->LiteralSuffix= "";
+      break;
+    }
+    /* else default */
+  case MYSQL_TYPE_DATE:
+  case MYSQL_TYPE_DATETIME:
+  case MYSQL_TYPE_NEWDATE:
+  case MYSQL_TYPE_TIME:
+  case MYSQL_TYPE_TIMESTAMP:
+  case MYSQL_TYPE_YEAR:
+    Record->LiteralPrefix="'";
+    Record->LiteralSuffix= "'";
+    break;
+  default:
+    Record->LiteralPrefix= "";
+    Record->LiteralSuffix= "";
+    break;
+  }
+
+  return 0;
+}
+/* }}} */
+
 /* {{{ MADB_DescSetIrdMetadata */
 my_bool 
 MADB_DescSetIrdMetadata(MADB_Stmt *Stmt, MYSQL_FIELD *Fields, unsigned int NumFields)
 {
-  MADB_Desc *Desc= Stmt->Ird;
-  MADB_DescRecord *Record;
-  SQLUINTEGER i;
+  SQLSMALLINT i;
 
-  for (i=0; i < NumFields; i++)
+  /* Perhaps we should call routine that does SQL_CLOSE here */
+  Stmt->Ird->Header.Count= 0;
+
+  for (i= 0; i < (SQLSMALLINT)NumFields; i++)
   {
-    if (!(Record= MADB_DescGetInternalRecord(Stmt->Ird, i, MADB_DESC_WRITE)))
+    if (MADB_SetIrdRecord(Stmt, MADB_DescGetInternalRecord(Stmt->Ird, i, MADB_DESC_WRITE), &Fields[i]))
+    {
       return 1;
-    
-    Record->CatalogName= Fields[i].catalog ? my_strdup(Fields[i].catalog, MYF(0)) : NULL;
-    Record->TableName= Fields[i].table ? my_strdup(Fields[i].table, MYF(0)) : NULL;
-    Record->ColumnName= Fields[i].name ? my_strdup(Fields[i].name, MYF(0)) : NULL;
-    Record->BaseTableName= Fields[i].org_table ? my_strdup(Fields[i].org_table, MYF(0)) : NULL;
-    Record->BaseColumnName= Fields[i].org_name ? my_strdup(Fields[i].org_name, MYF(0)) : NULL;
-    Record->AutoUniqueValue= (Fields[i].flags & AUTO_INCREMENT_FLAG) ? SQL_TRUE : SQL_FALSE;
-    Record->CaseSensitive= (Fields[i].flags & BINARY_FLAG) ? SQL_TRUE : SQL_FALSE;
-    Record->Nullable= ( (Fields[i].flags & NOT_NULL_FLAG) &&
-                        !Record->AutoUniqueValue &&
-                        Fields[i].type != MYSQL_TYPE_TIMESTAMP) ? SQL_NO_NULLS : SQL_NULLABLE;
-    Record->Unsigned= (Fields[i].flags & UNSIGNED_FLAG) ? SQL_TRUE : SQL_FALSE;
-    /* We assume it might be updatable if tablename exists */
-    Record->Updateable= (Fields[i].table && Fields[i].table[0]) ? SQL_ATTR_READWRITE_UNKNOWN : SQL_ATTR_READONLY;
-    /* 
-       SEARCHABLE:
-       Columns of type SQL_LONGVARCHAR and SQL_LONGVARBINARY usually return SQL_PRED_CHAR.
-    */
-    Record->Searchable= (Record->ConciseType == SQL_LONGVARCHAR ||
-                         Record->ConciseType == SQL_WLONGVARCHAR ||
-                         Record->ConciseType == SQL_LONGVARBINARY) ? SQL_PRED_CHAR : SQL_SEARCHABLE;
-    /*
-       RADIX:
-       If the data type in the SQL_DESC_TYPE field is an approximate numeric data type, this SQLINTEGER field 
-       contains a value of 2 because the SQL_DESC_PRECISION field contains the number of bits. 
-       If the data type in the SQL_DESC_TYPE field is an exact numeric data type, this field contains a value of 10 
-       because the SQL_DESC_PRECISION field contains the number of decimal digits. 
-       This field is set to 0 for all non-numeric data types.
-   */
-    switch (Fields[i].type) {
-    case MYSQL_TYPE_DECIMAL:
-    case MYSQL_TYPE_NEWDECIMAL:
-      Record->FixedPrecScale= SQL_FALSE;
-      Record->NumPrecRadix= 10;
-      Record->Precision= (SQLSMALLINT)Fields[i].length - 2;
-      Record->Scale= Fields[i].decimals;
-      break;
-    case MYSQL_TYPE_FLOAT:
-      Record->FixedPrecScale= SQL_FALSE;
-      Record->NumPrecRadix= 2;
-      Record->Precision= (SQLSMALLINT)Fields[i].length - 2;
-      //Record->Scale= Fields[i].decimals;
-      break;
-    case MYSQL_TYPE_BIT:
-      if (Fields[i].length > 1)
-        Record->Type= SQL_BINARY;
-      break;
-    case MYSQL_TYPE_DOUBLE:
-    case MYSQL_TYPE_TINY:
-    case MYSQL_TYPE_SHORT:
-    case MYSQL_TYPE_INT24:
-    case MYSQL_TYPE_YEAR:
-    case MYSQL_TYPE_LONG:
-    case MYSQL_TYPE_LONGLONG:
-      Record->NumPrecRadix= 10;
-      break;
-    default:
-      Record->NumPrecRadix= 0;
-      break;
-    }
-    Record->ConciseType= MADB_GetODBCType(&Fields[i]);
-    /* 
-       TYPE:
-       For the datetime and interval data types, this field returns the verbose data type: SQL_DATETIME or SQL_INTERVAL.
-    */
-    Record->Type= (Record->ConciseType ==  SQL_TYPE_DATE || Record->ConciseType ==  SQL_DATE ||
-                   Record->ConciseType ==  SQL_TYPE_TIME || Record->ConciseType ==  SQL_TIME ||
-                   Record->ConciseType ==  SQL_TYPE_TIMESTAMP || Record->ConciseType == SQL_TIMESTAMP) ?
-                     SQL_DATETIME : Record->ConciseType;
-
-    switch(Record->ConciseType) {
-    case SQL_TYPE_DATE:
-      Record->DateTimeIntervalCode= SQL_CODE_DATE;
-      break;
-    case SQL_TYPE_TIME:
-      Record->DateTimeIntervalCode= SQL_CODE_TIME;
-      break;
-    case SQL_TYPE_TIMESTAMP:
-      Record->DateTimeIntervalCode= SQL_CODE_TIMESTAMP;
-      break;
-    }
-    Record->DisplaySize= MADB_GetDisplaySize(Fields[i]);
-    Record->OctetLength= MADB_GetOctetLength(Fields[i], Stmt->Connection->mariadb->charset->char_maxlen);
-    Record->Length= MADB_GetDataSize(Record, Fields[i]);
-    
-    Record->TypeName= my_strdup(MADB_GetTypeName(Fields[i]), MYF(0));
-    switch(Fields[i].type) {
-    case MYSQL_TYPE_BLOB:
-    case MYSQL_TYPE_LONG_BLOB:
-    case MYSQL_TYPE_MEDIUM_BLOB:
-    case MYSQL_TYPE_STRING:
-    case MYSQL_TYPE_TINY_BLOB:
-    case MYSQL_TYPE_VAR_STRING:
-      if (Fields[i].flags & BINARY_FLAG)
-      {
-        Record->LiteralPrefix= "0x";
-        Record->LiteralSuffix= "";
-        break;
-      }
-      /* else default */
-    case MYSQL_TYPE_DATE:
-    case MYSQL_TYPE_DATETIME:
-    case MYSQL_TYPE_NEWDATE:
-    case MYSQL_TYPE_TIME:
-    case MYSQL_TYPE_TIMESTAMP:
-    case MYSQL_TYPE_YEAR:
-      Record->LiteralPrefix="'";
-      Record->LiteralSuffix= "'";
-      break;
-    default:
-      Record->LiteralPrefix= "";
-      Record->LiteralSuffix= "";
-      break;
     }
   }
   return 0;
 }
 /* }}} */
 
+/* {{{ MADB_FixOctetLength */
+void MADB_FixOctetLength(MADB_DescRecord *Record)
+{
+  switch (Record->ConciseType) {
+  case SQL_BIT:
+  case SQL_TINYINT:
+    Record->OctetLength= 1;
+    break;
+  case SQL_SMALLINT:
+    Record->OctetLength= 2;
+    break;
+  case SQL_INTEGER:
+  case SQL_REAL:
+    Record->OctetLength= 4;
+    break;
+  case SQL_BIGINT:
+  case SQL_DOUBLE:
+    Record->OctetLength= 8;
+    break;
+  case SQL_TYPE_DATE:
+    Record->OctetLength= sizeof(SQL_DATE_STRUCT);
+    break;
+  case SQL_TYPE_TIME:
+    Record->OctetLength= sizeof(SQL_TIME_STRUCT);
+    break;
+  case SQL_TYPE_TIMESTAMP:
+    Record->OctetLength= sizeof(SQL_TIMESTAMP_STRUCT);
+    break;
+  default:
+    Record->OctetLength= MIN(MADB_INT_MAX32, Record->OctetLength);
+  }
+}
+/* }}} */
+
+/* {{{ MADB_FixDisplayLength */
+void MADB_FixDisplaySize(MADB_DescRecord *Record, const MY_CHARSET_INFO *charset)
+{
+  switch (Record->ConciseType) {
+  case SQL_BIT:
+    Record->DisplaySize= 1;
+    break;
+  case SQL_TINYINT:
+    Record->DisplaySize= 4 - test(Record->Unsigned == SQL_TRUE);
+    break;
+  case SQL_SMALLINT:
+    Record->DisplaySize= 6 - test(Record->Unsigned == SQL_TRUE);
+    break;
+  case SQL_INTEGER:
+    Record->DisplaySize= 11 - test(Record->Unsigned == SQL_TRUE);
+    break;
+  case SQL_REAL:
+    Record->DisplaySize= 14;
+    break;
+  case SQL_BIGINT:
+    Record->DisplaySize= 20;
+    break;
+  case SQL_FLOAT:
+  case SQL_DOUBLE:
+    Record->DisplaySize= 24;
+    break;
+  case SQL_DECIMAL:
+  case SQL_NUMERIC:
+    Record->DisplaySize= Record->Precision + 2;
+    break;
+  case SQL_TYPE_DATE:
+    Record->DisplaySize= SQL_DATE_LEN;
+    break;
+  case SQL_TYPE_TIME:
+    Record->DisplaySize= SQL_TIME_LEN + MADB_FRACTIONAL_PART(Record->Scale);
+    break;
+  case SQL_TYPE_TIMESTAMP:
+    Record->DisplaySize= SQL_TIMESTAMP_LEN + MADB_FRACTIONAL_PART(Record->Scale);
+    break;
+  case SQL_BINARY:
+  case SQL_VARBINARY:
+  case SQL_LONGVARBINARY:
+    Record->DisplaySize=Record->OctetLength*2; /* For display in hex */
+    break;
+  case SQL_GUID:
+    Record->DisplaySize= 36;
+    break;
+  default:
+    if (charset == NULL || charset->mbmaxlen < 2/*i.e.0||1*/)
+    {
+      Record->DisplaySize=Record->OctetLength;
+    }
+    else
+    {
+      Record->DisplaySize=Record->OctetLength/charset->mbmaxlen;
+    }
+  }
+}
+/* }}} */
+
+/* {{{ MADB_FixDataSize - aka Column size */
+void MADB_FixDataSize(MADB_DescRecord *Record, const MY_CHARSET_INFO *charset)
+{
+  Record->Length= MADB_GetDataSize(Record->ConciseType, Record->OctetLength, Record->Unsigned == TRUE, Record->Precision, Record->Scale, charset->mbmaxlen );
+}
+/* }}} */
+
+/* {{{ MADB_FixIrdRecord */
+my_bool
+MADB_FixIrdRecord(MADB_Stmt *Stmt, MADB_DescRecord *Record)
+{
+  MY_CHARSET_INFO cs;
+
+  if (Record == NULL)
+  {
+    return 1;
+  }
+
+  MADB_FixOctetLength(Record);
+  /*
+      RADIX:
+      If the data type in the SQL_DESC_TYPE field is an approximate numeric data type, this SQLINTEGER field 
+      contains a value of 2 because the SQL_DESC_PRECISION field contains the number of bits. 
+      If the data type in the SQL_DESC_TYPE field is an exact numeric data type, this field contains a value of 10 
+      because the SQL_DESC_PRECISION field contains the number of decimal digits. 
+      This field is set to 0 for all non-numeric data types.
+  */
+  switch (Record->ConciseType) {
+  case SQL_DECIMAL:
+    Record->NumPrecRadix= 10;
+    Record->Precision= (SQLSMALLINT)Record->OctetLength - 2;
+    /*Record->Scale= Fields[i].decimals;*/
+    break;
+  case SQL_REAL:
+    /* Float*/
+    Record->NumPrecRadix= 2;
+    Record->Precision= (SQLSMALLINT)Record->OctetLength - 2;
+    break;
+  case SQL_DOUBLE:
+  case SQL_TINYINT:
+  case SQL_SMALLINT:
+  case SQL_INTEGER:
+  case SQL_BIGINT:
+    Record->NumPrecRadix= 10;
+    break;
+  default:
+    Record->NumPrecRadix= 0;
+    break;
+  }
+  /* 
+      TYPE:
+      For the datetime and interval data types, this field returns the verbose data type: SQL_DATETIME or SQL_INTERVAL.
+  */
+  Record->Type= (Record->ConciseType ==  SQL_TYPE_DATE || Record->ConciseType ==  SQL_DATE ||
+                  Record->ConciseType ==  SQL_TYPE_TIME || Record->ConciseType ==  SQL_TIME ||
+                  Record->ConciseType ==  SQL_TYPE_TIMESTAMP || Record->ConciseType == SQL_TIMESTAMP) ?
+                    SQL_DATETIME : Record->ConciseType;
+
+  switch(Record->ConciseType) {
+  case SQL_TYPE_DATE:
+    Record->DateTimeIntervalCode= SQL_CODE_DATE;
+    break;
+  case SQL_TYPE_TIME:
+    Record->DateTimeIntervalCode= SQL_CODE_TIME;
+    break;
+  case SQL_TYPE_TIMESTAMP:
+    Record->DateTimeIntervalCode= SQL_CODE_TIMESTAMP;
+    break;
+  }
+  /* 
+      SEARCHABLE:
+      Columns of type SQL_LONGVARCHAR and SQL_LONGVARBINARY usually return SQL_PRED_CHAR.
+  */
+  Record->Searchable= (Record->ConciseType == SQL_LONGVARCHAR ||
+                        Record->ConciseType == SQL_WLONGVARCHAR ||
+                        Record->ConciseType == SQL_LONGVARBINARY) ? SQL_PRED_CHAR : SQL_SEARCHABLE;
+
+  mariadb_get_infov(Stmt->Connection->mariadb, MARIADB_CONNECTION_MARIADB_CHARSET_INFO, (void*)&cs);
+
+  MADB_FixDisplaySize(Record, &cs);
+  MADB_FixDataSize(Record, &cs);
+    
+  switch(Record->ConciseType) {
+  case SQL_BINARY:
+  case SQL_VARBINARY:
+  case SQL_LONGVARBINARY:
+      Record->LiteralPrefix= "0x";
+      Record->LiteralSuffix= "";
+      break;
+  /* else default */
+  case SQL_TYPE_DATE:
+  case SQL_TYPE_TIME:
+  case SQL_TYPE_TIMESTAMP:
+    Record->LiteralPrefix="'";
+    Record->LiteralSuffix= "'";
+    break;
+  default:
+    Record->LiteralPrefix= "";
+    Record->LiteralSuffix= "";
+    break;
+  }
+
+  return 0;
+}
+/* }}} */
+
+/* {{{ MADB_FixColumnDataTypes */
+my_bool
+MADB_FixColumnDataTypes(MADB_Stmt *Stmt, MADB_ShortTypeInfo *ColTypesArr)
+{
+  SQLSMALLINT     i;
+  MADB_DescRecord *Record= NULL;
+
+  if (ColTypesArr == NULL)
+  {
+    return 1;
+  }
+  for (i=0; i < Stmt->Ird->Header.Count; ++i)
+  {
+    if (ColTypesArr[i].SqlType != 0)
+    {
+      Record= MADB_DescGetInternalRecord(Stmt->Ird, i, MADB_DESC_READ);
+
+      if (Record == NULL)
+      {
+        return 1;
+      }
+      Record->ConciseType= ColTypesArr[i].SqlType;
+      Record->Nullable= ColTypesArr[i].Nullable;
+
+      Record->Unsigned= ColTypesArr[i].Unsigned != 0 ? SQL_TRUE : SQL_FALSE;
+
+      if (ColTypesArr[i].OctetLength > 0)
+      {
+        Record->OctetLength= ColTypesArr[i].OctetLength;
+      }
+      if (MADB_FixIrdRecord(Stmt, Record))
+      {
+        return 1;
+      }
+    }
+  }
+
+  /* If the stmt is re-executed, we should be able to fix columns again */
+  Stmt->ColsTypeFixArr= ColTypesArr;
+  return 0;
+}
+/* }}} */
+
 void MADB_DescSetRecordDefaults(MADB_Desc *Desc, MADB_DescRecord *Record)
 {
   memset(Record, 0, sizeof(MADB_DescRecord));
@@ -289,22 +594,22 @@ void MADB_DescSetRecordDefaults(MADB_Desc *Desc, MADB_DescRecord *Record)
     Record->ConciseType= Record->Type= SQL_C_DEFAULT;
     break;
   case MADB_DESC_IPD:
-    Record->FixedPrecScale= SQL_TRUE;
+    Record->FixedPrecScale= SQL_FALSE;
     Record->LocalTypeName= "";
     Record->Nullable= SQL_NULLABLE;
     Record->ParameterType= SQL_PARAM_INPUT;
-    Record->TypeName= "VARCHAR";
+    MADB_RESET(Record->TypeName, "VARCHAR");
     Record->Unsigned= SQL_FALSE;
     Record->ColumnName= "";
     break;
   case MADB_DESC_IRD:
     Record->Nullable= SQL_NULLABLE_UNKNOWN;
-    Record->FixedPrecScale= SQL_TRUE;
+    Record->FixedPrecScale= SQL_FALSE;
     Record->CaseSensitive= SQL_TRUE;
     Record->ConciseType= SQL_VARCHAR;
     Record->AutoUniqueValue= SQL_FALSE;
     Record->Type= SQL_VARCHAR;
-    Record->TypeName= "VARCHAR";
+    MADB_RESET(Record->TypeName, "VARCHAR");
     Record->Unsigned= SQL_FALSE;
     break;
   }
@@ -312,10 +617,9 @@ void MADB_DescSetRecordDefaults(MADB_Desc *Desc, MADB_DescRecord *Record)
 /* }}} */
 
 /* {{{ MADB_DescGetInternalRecord */
-MADB_DescRecord *MADB_DescGetInternalRecord(MADB_Desc *Desc, SQLINTEGER RecordNumber, SQLSMALLINT Type)
+MADB_DescRecord *MADB_DescGetInternalRecord(MADB_Desc *Desc, SQLSMALLINT RecordNumber, SQLSMALLINT Type)
 {
   MADB_DescRecord *DescRecord;
-  SQLINTEGER Start= Desc->Records.elements;
 
   if (RecordNumber > (SQLINTEGER)Desc->Records.elements &&
       Type == MADB_DESC_READ)
@@ -326,7 +630,7 @@ MADB_DescRecord *MADB_DescGetInternalRecord(MADB_Desc *Desc, SQLINTEGER RecordNu
 
   while (RecordNumber >= (SQLINTEGER)Desc->Records.elements)
   {
-    if (!(DescRecord= (MADB_DescRecord *)alloc_dynamic(&Desc->Records)))
+    if (!(DescRecord= (MADB_DescRecord *)MADB_AllocDynamic(&Desc->Records)))
     {
       MADB_SetError(&Desc->Error, MADB_ERR_HY001, NULL, 0);
       return NULL;
@@ -336,8 +640,10 @@ MADB_DescRecord *MADB_DescGetInternalRecord(MADB_Desc *Desc, SQLINTEGER RecordNu
   }
 
   if (RecordNumber + 1 > Desc->Header.Count)
-    Desc->Header.Count= RecordNumber + 1;
+    Desc->Header.Count= (SQLSMALLINT)(RecordNumber + 1);
+
   DescRecord= ((MADB_DescRecord *)Desc->Records.buffer) + RecordNumber;
+
   return DescRecord;
 }
 /* }}} */
@@ -406,10 +712,10 @@ SQLRETURN MADB_DescGetField(SQLHDESC DescriptorHandle,
     *((SQLPOINTER *)ValuePtr)= (SQLPOINTER)Desc->Header.BindOffsetPtr;
     break;
   case SQL_DESC_BIND_TYPE:
-    *((SQLINTEGER *)ValuePtr)= Desc->Header.BindType;
+    *((SQLULEN *)ValuePtr)= Desc->Header.BindType;
     break;
   case SQL_DESC_COUNT:
-    *(SQLINTEGER *)ValuePtr= Desc->Header.Count;
+    *(SQLSMALLINT *)ValuePtr= Desc->Header.Count;
     break;
   case SQL_DESC_ROWS_PROCESSED_PTR:
     *((SQLPOINTER *)ValuePtr)= (SQLPOINTER)Desc->Header.RowsProcessedPtr;
@@ -418,22 +724,22 @@ SQLRETURN MADB_DescGetField(SQLHDESC DescriptorHandle,
     *((SQLINTEGER *)ValuePtr)= DescRecord->AutoUniqueValue;
     break;
   case SQL_DESC_BASE_COLUMN_NAME:
-    Length= MADB_SetString(isWChar ? CP_UTF8 : 0, ValuePtr, BufferLength, DescRecord->BaseColumnName, SQL_NTS, &Desc->Error);
+    Length= MADB_SetString(isWChar ? &utf8 : 0, ValuePtr, BufferLength, DescRecord->BaseColumnName, SQL_NTS, &Desc->Error);
     if (StringLengthPtr)
-      *StringLengthPtr= Length;
+      *StringLengthPtr= (SQLINTEGER)Length;
     break;
   case SQL_DESC_BASE_TABLE_NAME:
-    Length= MADB_SetString(isWChar ? CP_UTF8 : 0, ValuePtr, BufferLength, DescRecord->BaseTableName, SQL_NTS, &Desc->Error);
+    Length= MADB_SetString(isWChar ? &utf8 : 0, ValuePtr, BufferLength, DescRecord->BaseTableName, SQL_NTS, &Desc->Error);
     if (StringLengthPtr)
-      *StringLengthPtr= Length;
+      *StringLengthPtr= (SQLINTEGER)Length;
     break;
   case SQL_DESC_CASE_SENSITIVE:
     *((SQLINTEGER *)ValuePtr)= DescRecord->CaseSensitive;
     break;
   case SQL_DESC_CATALOG_NAME:
-    Length= MADB_SetString(isWChar ? CP_UTF8 : 0, ValuePtr, BufferLength, DescRecord->BaseCatalogName, SQL_NTS, &Desc->Error);
+    Length= MADB_SetString(isWChar ? &utf8 : 0, ValuePtr, BufferLength, DescRecord->BaseCatalogName, SQL_NTS, &Desc->Error);
     if (StringLengthPtr)
-      *StringLengthPtr= Length;
+      *StringLengthPtr= (SQLINTEGER)Length;
     break;
   case SQL_DESC_CONCISE_TYPE:
     *((SQLSMALLINT *)ValuePtr)= DescRecord->ConciseType;
@@ -448,7 +754,7 @@ SQLRETURN MADB_DescGetField(SQLHDESC DescriptorHandle,
     *((SQLINTEGER *)ValuePtr)= DescRecord->DateTimeIntervalPrecision;
     break;
   case SQL_DESC_FIXED_PREC_SCALE:
-    *((SQLINTEGER *)ValuePtr)= DescRecord->FixedPrecScale;
+    *((SQLSMALLINT *)ValuePtr)= DescRecord->FixedPrecScale;
     break;
   case SQL_DESC_INDICATOR_PTR:
     *((SQLPOINTER *)ValuePtr)= (SQLPOINTER)DescRecord->IndicatorPtr;
@@ -463,14 +769,14 @@ SQLRETURN MADB_DescGetField(SQLHDESC DescriptorHandle,
     *((SQLPOINTER *)ValuePtr)= (SQLPOINTER)DescRecord->LiteralSuffix;
     break;
   case SQL_DESC_LOCAL_TYPE_NAME:
-    Length= MADB_SetString(isWChar ? CP_UTF8 : 0, ValuePtr, BufferLength, DescRecord->LocalTypeName, SQL_NTS, &Desc->Error);
+    Length= MADB_SetString(isWChar ? &utf8 : 0, ValuePtr, BufferLength, DescRecord->LocalTypeName, SQL_NTS, &Desc->Error);
     if (StringLengthPtr)
-      *StringLengthPtr= Length;
+      *StringLengthPtr= (SQLINTEGER)Length;
     break;
   case SQL_DESC_NAME:
-    Length= MADB_SetString(isWChar ? CP_UTF8 : 0, ValuePtr, BufferLength, DescRecord->BaseColumnName, SQL_NTS, &Desc->Error);
+    Length= MADB_SetString(isWChar ? &utf8 : 0, ValuePtr, BufferLength, DescRecord->BaseColumnName, SQL_NTS, &Desc->Error);
     if (StringLengthPtr)
-      *StringLengthPtr= Length;
+      *StringLengthPtr= (SQLINTEGER)Length;
     DescRecord->Unnamed= SQL_NAMED;
     break;
   case SQL_DESC_NULLABLE:
@@ -480,43 +786,43 @@ SQLRETURN MADB_DescGetField(SQLHDESC DescriptorHandle,
     *((SQLINTEGER *)ValuePtr)= DescRecord->NumPrecRadix;
     break;
   case SQL_DESC_OCTET_LENGTH:
-    *((SQLINTEGER *)ValuePtr)= DescRecord->OctetLength;
+    *((SQLLEN *)ValuePtr)= DescRecord->OctetLength;
     break;
   case SQL_DESC_OCTET_LENGTH_PTR:
     *((SQLPOINTER *)ValuePtr)= (SQLPOINTER)DescRecord->OctetLengthPtr;
     break;
   case SQL_DESC_PARAMETER_TYPE:
-    *((SQLINTEGER *)ValuePtr)= DescRecord->ParameterType;
+    *((SQLSMALLINT *)ValuePtr)= DescRecord->ParameterType;
     break;
   case SQL_DESC_PRECISION:
     *((SQLINTEGER *)ValuePtr)= DescRecord->Precision;
     break;
 #if (ODBCVER >= 0x0350)
   case SQL_DESC_ROWVER:
-    *((SQLPOINTER *)ValuePtr)= (SQLPOINTER)DescRecord->RowVer;
+    *((SQLPOINTER *)ValuePtr)= (SQLPOINTER)(SQLULEN)DescRecord->RowVer;
     break;
 #endif
   case SQL_DESC_SCALE:
     *((SQLINTEGER *)ValuePtr)= DescRecord->Scale;
     break;
   case SQL_DESC_SCHEMA_NAME:
-    Length= MADB_SetString(isWChar ? CP_UTF8 : 0, ValuePtr, BufferLength, DescRecord->SchemaName, SQL_NTS, &Desc->Error);
+    Length= MADB_SetString(isWChar ? &utf8 : 0, ValuePtr, BufferLength, DescRecord->SchemaName, SQL_NTS, &Desc->Error);
     if (StringLengthPtr)
-      *StringLengthPtr= Length;
+      *StringLengthPtr= (SQLINTEGER)Length;
     break;
   case SQL_DESC_SEARCHABLE:
     *((SQLINTEGER *)ValuePtr)= DescRecord->Searchable;
     break;
   case SQL_DESC_TABLE_NAME:
-    Length= MADB_SetString(isWChar ? CP_UTF8 : 0, ValuePtr, BufferLength, DescRecord->TableName, SQL_NTS, &Desc->Error);
+    Length= MADB_SetString(isWChar ? &utf8 : 0, ValuePtr, BufferLength, DescRecord->TableName, SQL_NTS, &Desc->Error);
     if (StringLengthPtr)
-      *StringLengthPtr= Length;
+      *StringLengthPtr= (SQLINTEGER)Length;
     break;
   case SQL_DESC_TYPE:
     *((SQLINTEGER *)ValuePtr)= DescRecord->Type;
     break;
   case SQL_DESC_TYPE_NAME:
-    *StringLengthPtr= MADB_SetString(isWChar ? CP_UTF8 : 0, ValuePtr, BufferLength, DescRecord->TypeName, SQL_NTS, &Desc->Error);
+    *StringLengthPtr= (SQLINTEGER)MADB_SetString(isWChar ? &utf8 : 0, ValuePtr, BufferLength, DescRecord->TypeName, SQL_NTS, &Desc->Error);
      break;
   case SQL_DESC_UNSIGNED:
     *((SQLINTEGER *)ValuePtr)= DescRecord->Unsigned;
@@ -543,10 +849,9 @@ SQLRETURN MADB_DescSetField(SQLHDESC DescriptorHandle,
   ret= MADB_DeskCheckFldId(Desc, FieldIdentifier, MADB_DESC_WRITE);
 
   /* Application may set IPD's field SQL_DESC_UNNAMED to SQL_UNNAMED only */
-  if (FieldIdentifier == SQL_DESC_UNNAMED && (SQLSMALLINT)ValuePtr == SQL_NAMED)
+  if (FieldIdentifier == SQL_DESC_UNNAMED && (SQLSMALLINT)(SQLULEN)ValuePtr == SQL_NAMED)
   {
-    MADB_SetError(&Desc->Error, MADB_ERR_HY092, NULL, 0);
-    ret= Desc->Error.ReturnValue;
+    ret= MADB_SetError(&Desc->Error, MADB_ERR_HY092, NULL, 0);
   }
 
   if (!SQL_SUCCEEDED(ret))
@@ -555,19 +860,19 @@ SQLRETURN MADB_DescSetField(SQLHDESC DescriptorHandle,
   MADB_CLEAR_ERROR(&Desc->Error);
   switch (FieldIdentifier) {
   case SQL_DESC_ARRAY_SIZE:
-    Desc->Header.ArraySize= (SQLUINTEGER)ValuePtr;
+    Desc->Header.ArraySize= (SQLULEN)ValuePtr;
     return SQL_SUCCESS;
   case SQL_DESC_ARRAY_STATUS_PTR:
     Desc->Header.ArrayStatusPtr= (SQLUSMALLINT *)ValuePtr;
     return SQL_SUCCESS;
   case SQL_DESC_BIND_OFFSET_PTR:
-    Desc->Header.BindOffsetPtr= (SQLUINTEGER *)ValuePtr;
+    Desc->Header.BindOffsetPtr= (SQLULEN *)ValuePtr;
     return SQL_SUCCESS;
   case SQL_DESC_BIND_TYPE:
-    Desc->Header.BindType= (SQLUINTEGER)ValuePtr;
+    Desc->Header.BindType= (SQLINTEGER)(SQLLEN)ValuePtr;
     return SQL_SUCCESS;
   case SQL_DESC_COUNT:
-    Desc->Header.Count= (SQLINTEGER)ValuePtr;
+    Desc->Header.Count= (SQLSMALLINT)(SQLLEN)ValuePtr;
     return SQL_SUCCESS;
   case SQL_DESC_ROWS_PROCESSED_PTR:
     Desc->Header.RowsProcessedPtr= (SQLULEN *)ValuePtr;
@@ -581,47 +886,59 @@ SQLRETURN MADB_DescSetField(SQLHDESC DescriptorHandle,
 
     switch (FieldIdentifier) {
     case SQL_DESC_CONCISE_TYPE:
-      DescRecord->ConciseType= (SQLSMALLINT)ValuePtr;
+      DescRecord->ConciseType= (SQLSMALLINT)(SQLLEN)ValuePtr;
       DescRecord->Type= MADB_GetTypeFromConciseType(DescRecord->ConciseType);
+      if (DescRecord->Type == SQL_INTERVAL)
+      {
+        DescRecord->DateTimeIntervalCode= DescRecord->ConciseType - 100;
+      }
       break;
     case SQL_DESC_DATA_PTR:
       DescRecord->DataPtr= ValuePtr;
       break;
     case SQL_DESC_DATETIME_INTERVAL_CODE:
-      DescRecord->DateTimeIntervalCode= (SQLSMALLINT)ValuePtr;
+      DescRecord->DateTimeIntervalCode= (SQLSMALLINT)(SQLLEN)ValuePtr;
       break;
     case SQL_DESC_DATETIME_INTERVAL_PRECISION:
-      DescRecord->DateTimeIntervalPrecision= (SQLINTEGER)ValuePtr;
+      DescRecord->DateTimeIntervalPrecision= (SQLINTEGER)(SQLLEN)ValuePtr;
       break;
     case SQL_DESC_FIXED_PREC_SCALE:
-      DescRecord->FixedPrecScale= (SQLSMALLINT)ValuePtr;
+      DescRecord->FixedPrecScale= (SQLSMALLINT)(SQLLEN)ValuePtr;
       break;
     case SQL_DESC_INDICATOR_PTR:
-      DescRecord->IndicatorPtr= (SQLINTEGER *)ValuePtr;
+      DescRecord->IndicatorPtr= (SQLLEN *)ValuePtr;
       break;
     case SQL_DESC_LENGTH:
-      DescRecord->DescLength= (SQLINTEGER)ValuePtr;
+      DescRecord->DescLength= (SQLINTEGER)(SQLLEN)ValuePtr;
       break;
     case SQL_DESC_NUM_PREC_RADIX:
-      DescRecord->NumPrecRadix= (SQLINTEGER)ValuePtr;
+      DescRecord->NumPrecRadix= (SQLINTEGER)(SQLLEN)ValuePtr;
       break;
     case SQL_DESC_OCTET_LENGTH:
-      DescRecord->OctetLength= (SQLINTEGER)ValuePtr;
+      DescRecord->OctetLength= (SQLLEN)ValuePtr;
       break;
     case SQL_DESC_OCTET_LENGTH_PTR:
-      DescRecord->OctetLengthPtr= (SQLINTEGER *)ValuePtr;
+      DescRecord->OctetLengthPtr= (SQLLEN *)ValuePtr;
       break;
     case SQL_DESC_PARAMETER_TYPE:
-      DescRecord->ParameterType= (SQLSMALLINT)ValuePtr;
+      DescRecord->ParameterType= (SQLSMALLINT)(SQLLEN)ValuePtr;
       break;
     case SQL_DESC_PRECISION:
-      DescRecord->Precision= (SQLSMALLINT)ValuePtr;
+      DescRecord->Precision= (SQLSMALLINT)(SQLLEN)ValuePtr;
       break;
     case SQL_DESC_SCALE:
-      DescRecord->Scale= (SQLSMALLINT)ValuePtr;
+      if ((SQLSMALLINT)(SQLLEN)ValuePtr > MADB_MAX_SCALE)
+      {
+        DescRecord->Scale= MADB_MAX_SCALE;
+        ret= MADB_SetError(&Desc->Error, MADB_ERR_01S02, NULL, 0); 
+      }
+      else
+      {
+        DescRecord->Scale= (SQLSMALLINT)(SQLLEN)ValuePtr;
+      }
       break;
     case SQL_DESC_TYPE:
-      DescRecord->Type= (SQLSMALLINT)ValuePtr;
+      DescRecord->Type= (SQLSMALLINT)(SQLLEN)ValuePtr;
       DescRecord->ConciseType= DescRecord->Type;
       break;
     }
@@ -639,7 +956,9 @@ SQLRETURN MADB_DescSetField(SQLHDESC DescriptorHandle,
       break;
     }
   
-    if (DescRecord)
+    /* inUse is only used to check if column/parameter was bound or not. Thus we do not set it for each field, but only for those,
+       that make column/parameter "bound" */
+    if (DescRecord && (DescRecord->DataPtr != NULL || DescRecord->OctetLengthPtr != NULL || DescRecord->IndicatorPtr != NULL))
       DescRecord->inUse= 1;
   }
   return ret;
@@ -657,31 +976,47 @@ SQLRETURN MADB_DescCopyDesc(MADB_Desc *SrcDesc, MADB_Desc *DestDesc)
     MADB_SetError(&DestDesc->Error, MADB_ERR_HY016, NULL, 0);
     return SQL_ERROR;
   }
-  if (!SrcDesc->Header.Count)
+  if (SrcDesc->DescType == MADB_DESC_IRD && !SrcDesc->Header.Count)
   {
     MADB_SetError(&DestDesc->Error, MADB_ERR_HY007, NULL, 0);
     return SQL_ERROR;
   }
   /* make sure there aren't old records */
-  delete_dynamic(&DestDesc->Records);
-  if (my_init_dynamic_array(&DestDesc->Records, sizeof(MADB_DescRecord),
-                            SrcDesc->Records.elements, SrcDesc->Records.alloc_increment))
+  MADB_DeleteDynamic(&DestDesc->Records);
+  if (MADB_InitDynamicArray(&DestDesc->Records, sizeof(MADB_DescRecord),
+                            SrcDesc->Records.max_element, SrcDesc->Records.alloc_increment))
   {
     MADB_SetError(&DestDesc->Error, MADB_ERR_HY001, NULL, 0);
     return SQL_ERROR;
   }
 
   memcpy(&DestDesc->Header, &SrcDesc->Header, sizeof(MADB_Header));
-  DestDesc->AppType= SrcDesc->AppType;
+  
+  /* We don't copy AppType from Src to Dest. If we copy internal descriptor to the explicit/external, it stays explicit/external */
+
   DestDesc->DescType= SrcDesc->DescType;
   memcpy(&DestDesc->Error, &SrcDesc->Error, sizeof(MADB_Error));
 
   /* Since we never allocate pointers we can just copy content */
   memcpy(DestDesc->Records.buffer, SrcDesc->Records.buffer,
          SrcDesc->Records.size_of_element * SrcDesc->Records.max_element);
-  /* todo: internal buffer needs to be clearead or we need to move it outside of
-           record structure 
-  */
+  DestDesc->Records.elements= SrcDesc->Records.elements;
+
+  /* internal buffer needs to be clearead or we will get it freed twice with all nice subsequences */
+  {
+    unsigned int i;
+
+    for (i= 0; i < DestDesc->Records.elements; ++i)
+    {
+      MADB_DescRecord *Rec= MADB_DescGetInternalRecord(DestDesc, i, MADB_DESC_READ);
+
+      if (Rec != NULL)
+      {
+        Rec->InternalBuffer= NULL;
+      }
+    }
+  }
+
   return SQL_SUCCESS;
 }
 /* }}} */
@@ -711,7 +1046,7 @@ SQLRETURN MADB_DescGetRec(MADB_Desc *Desc,
   }
   
   /* SQL_DESC_NAME */
-  Length= MADB_SetString(isWChar ? CP_UTF8 : 0, Name, BufferLength, Record->BaseColumnName, SQL_NTS, &Desc->Error);
+  Length= MADB_SetString(isWChar ? &utf8 : 0, Name, BufferLength, Record->BaseColumnName, SQL_NTS, &Desc->Error);
   if (StringLengthPtr)
     *StringLengthPtr= (SQLSMALLINT)Length;
   Record->Unnamed= SQL_NAMED;
