@@ -30,8 +30,13 @@
 #define DSNKEY_DATABASE_INDEX 10
 #define DSNKEY_FP_INDEX       25
 #define DSNKEY_FPLIST_INDEX   26
+#define DSNKEY_STREAMRS_INDEX 43
 
 
+/* If adding new connstring option, one should add them to the end of the array section before aliases, because some arrays defining program logic
+   use indexes of an element in this array. In particular, in the setup lib there is array mapping dsn field(i.e. this array element) to parent window
+   and resource id of the contorl in the dialog. Thus if you still think it should go somewhere in the middle, mind also change that array and at least
+   DsnKeysSwitch below in this file, and index defines in this file, accordingly */
 MADB_DsnKey DsnKeys[]=
 {
   {"DSN",            offsetof(MADB_Dsn, DSNName),           DSN_TYPE_STRING, 0, 0}, /* 0 */
@@ -78,6 +83,12 @@ MADB_DsnKey DsnKeys[]=
   {"INTERACTIVE",    offsetof(MADB_Dsn, InteractiveClient), DSN_TYPE_BOOL,   0, 0},
   {"FORWARDONLY",    offsetof(MADB_Dsn, ForceForwardOnly),  DSN_TYPE_OPTION, MADB_OPT_FLAG_FORWARD_CURSOR, 0},
   {"SCHEMANOERROR",  offsetof(MADB_Dsn, NeglectSchemaParam),DSN_TYPE_BOOL,   0, 0},
+  {"READ_TIMEOUT",   offsetof(MADB_Dsn, ReadTimeout),       DSN_TYPE_INT,    0, 0},
+  {"WRITE_TIMEOUT",  offsetof(MADB_Dsn, WriteTimeout),      DSN_TYPE_INT,    0, 0}, /* 40 */
+  {"NOLOCALINFILE",  offsetof(MADB_Dsn, DisableLocalInfile),DSN_TYPE_BOOL,   0, 0},
+  {"NULLISCURRENT",  offsetof(MADB_Dsn, NullSchemaMeansCurrent),DSN_TYPE_BOOL, 0, 0},
+  {"STREAMRS",       offsetof(MADB_Dsn, StreamResult),      DSN_TYPE_OPTION, MADB_OPT_FLAG_NO_CACHE, 0},
+
   /* Aliases. Here offset is index of aliased key */
   {"SERVERNAME",     DSNKEY_SERVER_INDEX,                   DSN_TYPE_STRING, 0, 1},
   {"USER",           DSNKEY_UID_INDEX,                      DSN_TYPE_STRING, 0, 1},
@@ -85,13 +96,13 @@ MADB_DsnKey DsnKeys[]=
   {"DB",             DSNKEY_DATABASE_INDEX,                 DSN_TYPE_COMBO,  0, 1},
   {"SSLFP",          DSNKEY_FP_INDEX,                       DSN_TYPE_STRING, 0, 1},
   {"SSLFPLIST",      DSNKEY_FPLIST_INDEX,                   DSN_TYPE_STRING, 0, 1},
+  {"NO_CACHE",       DSNKEY_STREAMRS_INDEX,                 DSN_TYPE_BOOL,   0, 1},
 
   /* Terminating Null */
   {NULL, 0, DSN_TYPE_BOOL,0,0}
 };
 
-/* TODO: Shouln't 2nd be DSNKEY_OPTIONS_INDEX? But I seemed to remember it was removed for a reason... */
-#define IS_OPTIONS_BITMAP(key_index) (key_index == DSNKEY_OPTIONS_INDEX || key_index == DSNKEY_OPTIONS_INDEX)
+#define IS_OPTIONS_BITMAP(key_index) (key_index == DSNKEY_OPTIONS_INDEX)
 
 typedef struct
 {
@@ -125,8 +136,8 @@ MADB_Dsn *MADB_DSN_Init()
 
   if ((Dsn= (MADB_Dsn *)MADB_CALLOC(sizeof(MADB_Dsn))))
   {
-    Dsn->FreeMe= TRUE;
     Dsn->Keys= (MADB_DsnKey *)&DsnKeys;
+    /*Dsn->NullSchemaMeansCurrent = '\1';*/
   }
   return Dsn;
 }
@@ -162,8 +173,7 @@ void MADB_DSN_Free(MADB_Dsn *Dsn)
   MADB_FREE(Dsn->SaveFile);
   MADB_FREE(Dsn->ServerKey);
   MADB_FREE(Dsn->TlsKeyPwd);
-  if (Dsn->FreeMe)
-    MADB_FREE(Dsn); 
+  MADB_FREE(Dsn); 
 }
 /* }}} */
 
